# baudrates for GMC device
# First one is most common default baudrate, second one is second most common default.
# The rest are not a default but they are an option on the device.
# Is a lower baudrate more stable?
BAUDRATES = (
    115200,
    57600,
    38400,
    28800,
    19200,
    14400,
    9600,
    4800,
    2400,
    1200,
)
